import logging
import string
import operator
from collections import deque
from decimal import Decimal


logger = logging.getLogger(__name__)


is_digit = lambda ch: ch and ch in string.digits


TOKEN_NUM = 0
TOKEN_OP = 1
TOKEN_LPAR = 2
TOKEN_RPAR = 3
TOKEN_IDEN = 4
TOKEN_COMMA = 5


class Node(object):
    def value(self):
        raise NotImplementedError


class NumNode(Node):
    def __init__(self, num):
        self.num = num

    def value(self):
        return Decimal(self.num)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, repr(self.num))


class UnaryNode(Node):
    def __init__(self, op_sign, node):
        self.op_sign = op_sign
        self.node = node

    def value(self):
        if self.op_sign == '-':
            return -self.node.value()

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__, 
            self.op_sign, 
            repr(self.node)
        )


class BinOpNode(Node):
    def __init__(self, n1, n2, op):
        self.n1 = n1
        self.n2 = n2
        self.op = op

    def value(self):
        return self.op(self.n1.value(), self.n2.value())

    def __repr__(self):
        return '{}({}, {}, {})'.format(
            self.__class__.__name__,
            self.op.__name__, 
            repr(self.n1), 
            repr(self.n2)
        )


class FuncNode(Node):
    def __init__(self, func, *nodes):
        self.func = func
        self.nodes = nodes

    def value(self):
        return self.func(*[n.value() for n in self.nodes])

    def __repr__(self):
        return 'FuncNode({}, {})'.format(
            self.func.__name__,
            ', '.join([repr(n) for n in self.nodes])
        )


class Lexer(object):
    def __init__(self, expr):
        self.expr = list(expr)
        self.expr.append(None)
        self.idx = 0
        self.queue = deque()

    def peek_token(self):
        if self.queue:
            return self.queue[0]

        token = self.pop_token()
        if token[0] is not None:
            self.queue.append(token)
            return token

        return None, None

    def pop_token(self):
        if self.queue:
            return self.queue.popleft()

        while self.expr[self.idx] in {' ', '\t'}:
            self.idx += 1

        if self.expr[self.idx] == None:
            return None, None

        if is_digit(self.expr[self.idx]):
            start = self.idx
            while is_digit(self.expr[self.idx+1]):
                self.idx += 1
            if self.expr[self.idx+1] == '.':
                self.idx += 1
            while is_digit(self.expr[self.idx+1]):
                self.idx += 1
            self.idx += 1
            return ''.join(self.expr[start:self.idx]), TOKEN_NUM
        if self.expr[self.idx] == ',':
            self.idx += 1
            return ',', TOKEN_COMMA
        if self.expr[self.idx] == '(':
            self.idx += 1
            return '(', TOKEN_LPAR
        if self.expr[self.idx] == ')':
            self.idx += 1
            return ')', TOKEN_RPAR
        if self.expr[self.idx] in {'+', '-', '*', '/'}:
            ch = self.expr[self.idx]
            self.idx += 1
            return ch, TOKEN_OP
        if self.expr[self.idx] in string.ascii_letters:
            start = self.idx
            while self.expr[self.idx] in string.ascii_letters:
                self.idx += 1
            return ''.join(self.expr[start:self.idx]), TOKEN_IDEN

        return None, None


OPERATOR_FUNC_MAP = {
    '+': operator.add,
    '-': operator.sub,
    '*': operator.mul,
    '/': operator.div,
}


from plz.funcs import statistics


FUNCTIONS = {
    'mean': statistics.mean,
    'median': statistics.median,
}


class ParseError(Exception):
    pass


class Parser(object):
    """

    BNF:
        expr  = add
        add   = mul ((+|-) mul)*
        mul   = unary ((*|/) unary)*
        unary = [+/-]? prim
        prim  = NAME ( list ) | NUM
        list  = expr (, expr)*
    """
    def __init__(self, expr):
        self.expr = expr
        self.lexer = Lexer(expr)

    def error(self, msg, num=None):
        logger.debug('\n'.join([
            self.expr,
            (' ' * (num - 2)) + '"',
            msg,
        ]))
        raise ParseError(msg)

    def parse_mul(self):
        left = self.parse_unary()
        op, __ = self.lexer.peek_token()
        while op in {'*', '/'}:
            self.lexer.pop_token()
            right = self.parse_unary()
            left = BinOpNode(left, right, OPERATOR_FUNC_MAP[op])
            op, __ = self.lexer.peek_token()
        return left

    def parse_add(self):
        left = self.parse_mul()
        op, __ = self.lexer.peek_token()
        while op in {'+', '-'}:
            self.lexer.pop_token()
            right = self.parse_mul()
            left = BinOpNode(left, right, OPERATOR_FUNC_MAP[op])
            op, __ = self.lexer.peek_token()
        return left

    def parse_unary(self):
        t, t_type = self.lexer.peek_token()
        if t_type == TOKEN_OP:
            self.lexer.pop_token()
            return UnaryNode(t, self.parse_unary())
        return self.parse_prim()

    def parse_prim(self):
        t, t_type = self.lexer.peek_token()
        if t_type == TOKEN_IDEN:
            return self.parse_func()
        if t_type == TOKEN_LPAR:
            self.lexer.pop_token()
            n = self.parse_add()
            t, t_type = self.lexer.pop_token()
            if t_type != TOKEN_RPAR:
                raise ParseError('prim')
            return n
        return self.parse_num()

    def parse_func(self):
        func_name, t_type = self.lexer.pop_token()
        if t_type != TOKEN_IDEN:
            raise ParseError('func')
        t, t_type = self.lexer.pop_token()
        if t_type != TOKEN_LPAR:
            import pdb; pdb.set_trace()
            raise ParseError('func')
        
        args = self.parse_list()
        return FuncNode(FUNCTIONS[func_name], *args)

    def parse_list(self):
        args = []
        args.append(self.parse_add())
        t, t_type = self.lexer.peek_token()
        if t_type == TOKEN_COMMA:
            self.lexer.pop_token()
            args += self.parse_list()
        return args

    def parse_num(self):
        t, t_type = self.lexer.pop_token()
        if t_type != TOKEN_NUM:
            self.error('expected number', self.lexer.idx - len(t or ''))
        return NumNode(t)

    def parse_expr(self):
        return self.parse_add()

    def parse(self):
        return self.parse_expr()


def eval(expr):
    root = Parser(expr).parse()
    return root.value()
